import aiohttp

from . import fetch_nodeinfo


async def test_fetch_nodeinfo():
    async with aiohttp.ClientSession() as session:
        result = await fetch_nodeinfo(session, "mymath.rocks")

        assert result

        assert result["software"]["name"] == "bovine"
        assert "version" in result["software"]



from urllib.parse import urlparse


def netloc(identifier: str | None) -> str | None:
    if identifier is None:
        return None
    if identifier.startswith("http://") or identifier.startswith("https"):
        return urlparse(identifier).netloc
    return None


class BovineObject:
    """Initializes an Object to be parsed

    :param data: The data to be parsed. It is assumed that data is
        compacted against the **about:bovine** context and the parts
        making it up are validated."""

    def __init__(self, data: dict, domain: str = None, domain_may_differ: bool = False):
        if data.get("@context", "about:bovine") != "about:bovine":
            raise ValueError("@context must be 'about:bovine'")
        self.data = {"@context": "about:bovine", **data}

        if domain:
            id_domain = netloc(data.get("id"))
            if domain_may_differ:
                if id_domain:
                    self.domain = id_domain
                else:
                    self.domain = domain
            else:
                if id_domain and domain != id_domain:
                    raise ValueError("Domain and netloc of id must match")
                self.domain = domain
        else:
            if data.get("id") is None:
                raise ValueError(
                    "Either domain needs to be specified or data needs to have an id"
                )
            self.domain = urlparse(data.get("id")).netloc

    @property
    def actor_id(self):
        """If the object has an actor property, set to its id"""
        result = self.data.get("actor")
        if isinstance(result, dict):
            result = result.get("id")
        if netloc(result) and netloc(result) != self.domain:
            raise ValueError("actor on different domain than object")
        return result

    @property
    def object_id(self):
        """If the object has an object property, set to its id"""
        result = self.data.get("object")
        if isinstance(result, dict):
            return result.get("id")
        return result



from typing import Optional
from urllib.parse import urlparse


def actor_for_object(data: dict) -> str:
    """Look up for the actor id either from attributedTo or actor"""
    if "attributedTo" in data:
        actor = data.get("attributedTo")
    else:
        actor = data.get("actor")
    actor = id_for_object(actor)

    if actor:
        return actor

    return "__NO__ACTOR__"


def id_for_object(data: Optional[dict | str]) -> Optional[str]:
    """Determines the id of an object"""

    if data is None:
        return None
    if isinstance(data, str):
        return data
    return data.get("id", None)


def property_for_key_as_set(data, key):
    if data is None:
        return set()
    result = data.get(key, [])
    if isinstance(result, str):
        return set([result])
    return set(result)


def recipients_for_object(data: dict) -> set:
    """Combines the recipients from to, cc, bto, bcc, audience into a set"""
    return set.union(
        *[
            property_for_key_as_set(data, key)
            for key in ["to", "cc", "bto", "bcc", "audience"]
        ]
    )


def remove_public(recipients):
    return {
        x
        for x in recipients
        if x
        not in ["Public", "as:Public", "https://www.w3.org/ns/activitystreams#Public"]
    }


def is_public(data: dict) -> bool:
    """Determines if the object should be considered public based on its recipients"""
    recipients = recipients_for_object(data)

    return any(
        x in recipients
        for x in ["Public", "as:Public", "https://www.w3.org/ns/activitystreams#Public"]
    )


def fediverse_handle_from_actor(actor: dict) -> str:
    """Given an actor object, i.e. a dict, determines the fediverse handle"""
    host = urlparse(actor["id"]).netloc
    username = urlparse(actor["id"]).path.split("/")[-1]

    if "preferredUsername" in actor:
        username = actor["preferredUsername"]

    return f"{username}@{host}"


def copy_to_and_cc(origin: dict, destination: dict) -> dict:
    """Copies the audience from the origin object to the destination object"""

    for key in ["to", "cc", "bto", "bcc", "audience"]:
        if key in origin:
            destination[key] = list(
                property_for_key_as_set(origin, key)
                | property_for_key_as_set(destination, key)
            )

    return destination



from .types import ServerSentEvent


def test_server_sent_event():
    normalized_string = """data: {"json":true}
event: outbox

""".encode(
        "utf-8"
    )

    sse = ServerSentEvent.parse(normalized_string)

    result = sse.encode()

    assert result == normalized_string



import json
import logging
from typing import List

import requests_cache
from pyld import jsonld

from bovine.utils.pyld_requests import requests_document_loader

requests_cache.install_cache("context_cache")

logger = logging.getLogger(__name__)


default_context = [
    "https://www.w3.org/ns/activitystreams",
    "https://w3id.org/security/v1",
    {
        "Hashtag": "as:Hashtag",
    },
]
"""Defines the context used to communicate with other Fediverse software"""


bovine_context = [
    "https://www.w3.org/ns/activitystreams",
    {
        "publicKey": {"@id": "https://w3id.org/security#publicKey", "@type": "@id"},
        "publicKeyPem": "https://w3id.org/security#publicKeyPem",
        "owner": {"@id": "https://w3id.org/security#owner", "@type": "@id"},
        "to": {"@id": "as:to", "@type": "@id", "@container": "@set"},
        "cc": {"@id": "as:cc", "@type": "@id", "@container": "@set"},
        "tag": {"@id": "as:tag", "@type": "@id", "@container": "@set"},
        "items": {"@id": "as:items", "@type": "@id", "@container": "@set"},
        "attachment": {"@id": "as:attachment", "@type": "@id", "@container": "@set"},
        "Hashtag": "as:Hashtag",
    },
]
""" Defines the context about:bovine used internally in the bovine stack"""

bovine_context_name = "about:bovine"
"""Defines the name of the bovine context"""

activitystreams_context = {
    "@vocab": "_:",
    "Accept": "as:Accept",
    "Activity": "as:Activity",
    "Add": "as:Add",
    "Announce": "as:Announce",
    "Application": "as:Application",
    "Arrive": "as:Arrive",
    "Article": "as:Article",
    "Audio": "as:Audio",
    "Block": "as:Block",
    "Collection": "as:Collection",
    "CollectionPage": "as:CollectionPage",
    "Create": "as:Create",
    "Delete": "as:Delete",
    "Dislike": "as:Dislike",
    "Document": "as:Document",
    "Event": "as:Event",
    "Flag": "as:Flag",
    "Follow": "as:Follow",
    "Group": "as:Group",
    "Ignore": "as:Ignore",
    "Image": "as:Image",
    "IntransitiveActivity": "as:IntransitiveActivity",
    "Invite": "as:Invite",
    "IsContact": "as:IsContact",
    "IsFollowedBy": "as:IsFollowedBy",
    "IsFollowing": "as:IsFollowing",
    "IsMember": "as:IsMember",
    "Join": "as:Join",
    "Leave": "as:Leave",
    "Like": "as:Like",
    "Link": "as:Link",
    "Listen": "as:Listen",
    "Mention": "as:Mention",
    "Move": "as:Move",
    "Note": "as:Note",
    "Object": "as:Object",
    "Offer": "as:Offer",
    "OrderedCollection": "as:OrderedCollection",
    "OrderedCollectionPage": "as:OrderedCollectionPage",
    "Organization": "as:Organization",
    "Page": "as:Page",
    "Person": "as:Person",
    "Place": "as:Place",
    "Profile": "as:Profile",
    "Public": {"@id": "as:Public", "@type": "@id"},
    "Question": "as:Question",
    "Read": "as:Read",
    "Reject": "as:Reject",
    "Relationship": "as:Relationship",
    "Remove": "as:Remove",
    "Service": "as:Service",
    "TentativeAccept": "as:TentativeAccept",
    "TentativeReject": "as:TentativeReject",
    "Tombstone": "as:Tombstone",
    "Travel": "as:Travel",
    "Undo": "as:Undo",
    "Update": "as:Update",
    "Video": "as:Video",
    "View": "as:View",
    "accuracy": {"@id": "as:accuracy", "@type": "xsd:float"},
    "actor": {"@id": "as:actor", "@type": "@id"},
    "alsoKnownAs": {"@id": "as:alsoKnownAs", "@type": "@id"},
    "altitude": {"@id": "as:altitude", "@type": "xsd:float"},
    "anyOf": {"@id": "as:anyOf", "@type": "@id"},
    "as": "https://www.w3.org/ns/activitystreams#",
    "attachment": {"@id": "as:attachment", "@type": "@id"},
    "attributedTo": {"@id": "as:attributedTo", "@type": "@id"},
    "audience": {"@id": "as:audience", "@type": "@id"},
    "bcc": {"@id": "as:bcc", "@type": "@id"},
    "bto": {"@id": "as:bto", "@type": "@id"},
    "cc": {"@id": "as:cc", "@type": "@id"},
    "closed": {"@id": "as:closed", "@type": "xsd:dateTime"},
    "content": "as:content",
    "contentMap": {"@container": "@language", "@id": "as:content"},
    "context": {"@id": "as:context", "@type": "@id"},
    "current": {"@id": "as:current", "@type": "@id"},
    "deleted": {"@id": "as:deleted", "@type": "xsd:dateTime"},
    "describes": {"@id": "as:describes", "@type": "@id"},
    "duration": {"@id": "as:duration", "@type": "xsd:duration"},
    "endTime": {"@id": "as:endTime", "@type": "xsd:dateTime"},
    "endpoints": {"@id": "as:endpoints", "@type": "@id"},
    "first": {"@id": "as:first", "@type": "@id"},
    "followers": {"@id": "as:followers", "@type": "@id"},
    "following": {"@id": "as:following", "@type": "@id"},
    "formerType": {"@id": "as:formerType", "@type": "@id"},
    "generator": {"@id": "as:generator", "@type": "@id"},
    "height": {"@id": "as:height", "@type": "xsd:nonNegativeInteger"},
    "href": {"@id": "as:href", "@type": "@id"},
    "hreflang": "as:hreflang",
    "icon": {"@id": "as:icon", "@type": "@id"},
    "id": "@id",
    "image": {"@id": "as:image", "@type": "@id"},
    "inReplyTo": {"@id": "as:inReplyTo", "@type": "@id"},
    "inbox": {"@id": "ldp:inbox", "@type": "@id"},
    "instrument": {"@id": "as:instrument", "@type": "@id"},
    "items": {"@id": "as:items", "@type": "@id"},
    "last": {"@id": "as:last", "@type": "@id"},
    "latitude": {"@id": "as:latitude", "@type": "xsd:float"},
    "ldp": "http://www.w3.org/ns/ldp#",
    "liked": {"@id": "as:liked", "@type": "@id"},
    "likes": {"@id": "as:likes", "@type": "@id"},
    "location": {"@id": "as:location", "@type": "@id"},
    "longitude": {"@id": "as:longitude", "@type": "xsd:float"},
    "mediaType": "as:mediaType",
    "name": "as:name",
    "nameMap": {"@container": "@language", "@id": "as:name"},
    "next": {"@id": "as:next", "@type": "@id"},
    "oauthAuthorizationEndpoint": {
        "@id": "as:oauthAuthorizationEndpoint",
        "@type": "@id",
    },
    "oauthTokenEndpoint": {"@id": "as:oauthTokenEndpoint", "@type": "@id"},
    "object": {"@id": "as:object", "@type": "@id"},
    "oneOf": {"@id": "as:oneOf", "@type": "@id"},
    "orderedItems": {"@container": "@list", "@id": "as:items", "@type": "@id"},
    "origin": {"@id": "as:origin", "@type": "@id"},
    "outbox": {"@id": "as:outbox", "@type": "@id"},
    "partOf": {"@id": "as:partOf", "@type": "@id"},
    "preferredUsername": "as:preferredUsername",
    "prev": {"@id": "as:prev", "@type": "@id"},
    "preview": {"@id": "as:preview", "@type": "@id"},
    "provideClientKey": {"@id": "as:provideClientKey", "@type": "@id"},
    "proxyUrl": {"@id": "as:proxyUrl", "@type": "@id"},
    "published": {"@id": "as:published", "@type": "xsd:dateTime"},
    "radius": {"@id": "as:radius", "@type": "xsd:float"},
    "rel": "as:rel",
    "relationship": {"@id": "as:relationship", "@type": "@id"},
    "replies": {"@id": "as:replies", "@type": "@id"},
    "result": {"@id": "as:result", "@type": "@id"},
    "sharedInbox": {"@id": "as:sharedInbox", "@type": "@id"},
    "shares": {"@id": "as:shares", "@type": "@id"},
    "signClientKey": {"@id": "as:signClientKey", "@type": "@id"},
    "source": "as:source",
    "startIndex": {"@id": "as:startIndex", "@type": "xsd:nonNegativeInteger"},
    "startTime": {"@id": "as:startTime", "@type": "xsd:dateTime"},
    "streams": {"@id": "as:streams", "@type": "@id"},
    "subject": {"@id": "as:subject", "@type": "@id"},
    "summary": "as:summary",
    "summaryMap": {"@container": "@language", "@id": "as:summary"},
    "tag": {"@id": "as:tag", "@type": "@id"},
    "target": {"@id": "as:target", "@type": "@id"},
    "to": {"@id": "as:to", "@type": "@id"},
    "totalItems": {"@id": "as:totalItems", "@type": "xsd:nonNegativeInteger"},
    "type": "@type",
    "units": "as:units",
    "updated": {"@id": "as:updated", "@type": "xsd:dateTime"},
    "uploadMedia": {"@id": "as:uploadMedia", "@type": "@id"},
    "url": {"@id": "as:url", "@type": "@id"},
    "vcard": "http://www.w3.org/2006/vcard/ns#",
    "width": {"@id": "as:width", "@type": "xsd:nonNegativeInteger"},
    "xsd": "http://www.w3.org/2001/XMLSchema#",
}


def wrapper(url, options, **kwargs):
    if url == bovine_context_name:
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "documentUrl": url,
            "document": {"@context": bovine_context},
        }
    elif url == "https://www.w3.org/ns/activitystreams":
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "documentUrl": url,
            "document": {"@context": activitystreams_context},
        }
    elif url.startswith("http://joinmastodon.org/ns"):
        # See https://github.com/go-fed/activity/issues/152 for why
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "documentUrl": url,
            "document": {"@context": {}},
        }
    elif url == "http://schema.org":
        url = "https://schema.org/docs/jsonldcontext.jsonld"
    result = requests_document_loader(timeout=60)(url, options)
    return result


jsonld.set_document_loader(wrapper)


async def split_into_objects(input_data: dict) -> List[dict]:
    """Takes an object with an "id" property and separates
    out all the subobjects with an id"""

    if "@context" not in input_data:
        logger.warning("@context missing in %s", json.dumps(input_data))
        input_data["@context"] = default_context

    context = input_data["@context"]
    flattened = jsonld.flatten(input_data)
    compacted = jsonld.compact(flattened, context)

    if "@graph" not in compacted:
        return [compacted]

    local, remote = split_remote_local(compacted["@graph"])

    return [frame_object(obj, local, context) for obj in remote]


def frame_object(obj: dict, local: List[dict], context) -> dict:
    to_frame = {"@context": context, "@graph": [obj] + local}
    frame = {"@context": context, "id": obj["id"]}
    return jsonld.frame(to_frame, frame)


def split_remote_local(graph):
    local = [x for x in graph if x["id"].startswith("_")]
    remote = [x for x in graph if not x["id"].startswith("_")]

    return local, remote


def combine_items(data: dict, items: List[dict]) -> dict:
    """Takes data and replaces ids by the corresponding objects from items"""
    return frame_object(data, items, data["@context"])


def with_bovine_context(data: dict) -> dict:
    """Returns the object with the about:bovine context"""
    return use_context(data, "about:bovine")


def with_external_context(data: dict) -> dict:
    """Returns the object with the default external context"""
    return use_context(data, default_context)


def with_activitystreams_context(data: dict) -> dict:
    """Returns the object with the ActivityStreams context"""
    return use_context(data, "https://www.w3.org/ns/activitystreams")


def use_context(data, context):
    return jsonld.compact(data, context)


def value_from_object(data, key):
    result = data.get(key)
    if result is None:
        return result
    if isinstance(result, str):
        return result
    return result["@value"]



import base64
import hashlib
import logging

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, padding, rsa
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)
from multiformats import multibase, multicodec

logger = logging.getLogger(__name__)


def content_digest_sha256(content: str | bytes) -> str:
    """Computes the SHA256 digest of given content"""
    if isinstance(content, str):
        content = content.encode("utf-8")

    digest = base64.standard_b64encode(hashlib.sha256(content).digest()).decode("utf-8")
    return "sha-256=" + digest


def sign_message(private_key, message):
    try:
        key = load_pem_private_key(private_key.encode("utf-8"), password=None)
        assert isinstance(key, rsa.RSAPrivateKey)
    except Exception as e:
        logger.error(e)
        logger.error(private_key)
        raise (e)

    return base64.standard_b64encode(
        key.sign(
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    ).decode("utf-8")


def verify_signature(public_key, message, signature):
    public_key_loaded = load_pem_public_key(public_key.encode("utf-8"))

    assert isinstance(public_key_loaded, rsa.RSAPublicKey)

    try:
        public_key_loaded.verify(
            base64.standard_b64decode(signature),
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    except InvalidSignature:
        logger.warning("invalid signature")
        return False

    return True


def public_key_to_did_key(public_key: ed25519.Ed25519PublicKey) -> str:
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw,
    )

    wrapped = multicodec.wrap("ed25519-pub", public_bytes)
    encoded = multibase.encode(wrapped, "base58btc")

    return "did:key:" + encoded


def did_key_to_public_key(did: str) -> ed25519.Ed25519PublicKey:
    assert did.startswith("did:key:")
    decoded = multibase.decode(did[8:])
    codec, key_bytes = multicodec.unwrap(decoded)
    assert codec.name == "ed25519-pub"

    return ed25519.Ed25519PublicKey.from_public_bytes(key_bytes)


def private_key_to_ed25519(private_key_str: str) -> ed25519.Ed25519PrivateKey:
    decoded = multibase.decode(private_key_str)
    codec, key_bytes = multicodec.unwrap(decoded)
    assert codec.name == "ed25519-priv"

    return ed25519.Ed25519PrivateKey.from_private_bytes(key_bytes)



import logging

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ed25519
from multiformats import multibase, multicodec

from .helper import did_key_to_public_key, sign_message, verify_signature

logger = logging.getLogger(__name__)


def build_signature(host, method, target):
    return (
        HttpSignature()
        .with_field("(request-target)", f"{method} {target}")
        .with_field("host", host)
    )


class HttpSignature:
    """Helper class to build http signatures

    Usage: Add fields used for signature with `with_fields`. Then
    use `build_signature` or `verify` depending on use case.
    """

    def __init__(self):
        self.fields = []

    def build_signature(self, key_id: str, private_key: str):
        """Returns the signature string when signed with private_key"""
        message = self.build_message()

        signature_string = sign_message(private_key, message)
        headers = " ".join(name for name, _ in self.fields)

        signature_parts = [
            f'keyId="{key_id}"',
            'algorithm="rsa-sha256"',  # FIXME: Should other algorithms be supported?
            f'headers="{headers}"',
            f'signature="{signature_string}"',
        ]

        return ",".join(signature_parts)

    def ed25519_sign(self, private_encoded):
        private_bytes = multicodec.unwrap(multibase.decode(private_encoded))[1]
        private_key = ed25519.Ed25519PrivateKey.from_private_bytes(private_bytes)

        message = self.build_message()

        return multibase.encode(private_key.sign(message.encode("utf-8")), "base58btc")

    def ed25519_verify(self, did_key, signature):
        public_key = did_key_to_public_key(did_key)

        signature = multibase.decode(signature)

        message = self.build_message().encode("utf-8")

        try:
            public_key.verify(signature, message)
        except InvalidSignature:
            return False

        return True

    def verify(self, public_key: str, signature: str):
        """Verifies signature"""
        message = self.build_message()
        return verify_signature(public_key, message, signature)

    def build_message(self):
        """Builds the message"""
        return "\n".join(f"{name}: {value}" for name, value in self.fields)

    def with_field(self, field_name, field_value):
        """Adds a field to be used when building a http signature"""
        self.fields.append((field_name, field_value))
        return self

    @property
    def headers(self):
        """Headers as specified when building http signature"""
        return {name: value for name, value in self.fields if name[0] != "("}



from bovine.activitystreams.object_factory import Object

from . import (
    actor_for_object,
    fediverse_handle_from_actor,
    is_public,
    recipients_for_object,
)


def test_get_recipients():
    note = Object(
        type="Note", attributed_to="account", followers="account/followers"
    ).as_public()
    note.cc.add("cc")
    note.to.add("to")
    note.to.add("same")
    note.cc.add("same")
    note = note.build()

    recipients = recipients_for_object(note)

    assert recipients == {
        "account/followers",
        "https://www.w3.org/ns/activitystreams#Public",
        "cc",
        "to",
        "same",
    }


def test_is_public():
    note = (
        Object(type="Note", attributed_to="account", followers="account/followers")
        .as_public()
        .build()
    )
    assert is_public(note)

    note = Object(
        type="Note", attributed_to="account", followers="account/followers"
    ).build()
    assert not is_public(note)

    note = Object(type="Note", attributed_to="account", followers="account/followers")
    note.cc.add("someone")
    note = note.build()
    assert not is_public(note)

    note = (
        Object(type="Note", attributed_to="account", followers="account/followers")
        .as_unlisted()
        .build()
    )
    assert is_public(note)

    note = Object(
        type="Note", attributed_to="account", followers="account/followers"
    ).build()
    note["to"] = "as:Public"
    assert is_public(note)


def test_actor_for_object():
    assert actor_for_object({}) == "__NO__ACTOR__"
    assert actor_for_object({"actor": "alice"}) == "alice"
    assert actor_for_object({"actor": {"id": "alice"}}) == "alice"
    assert actor_for_object({"actor": {}}) == "__NO__ACTOR__"
    assert actor_for_object({"attributedTo": "alice"}) == "alice"
    assert actor_for_object({"attributedTo": {"id": "alice"}}) == "alice"
    assert actor_for_object({"attributedTo": {}}) == "__NO__ACTOR__"


def test_fediverse_handle():
    assert (
        fediverse_handle_from_actor({"id": "https://abel/actor/alice"}) == "alice@abel"
    )
    assert (
        fediverse_handle_from_actor(
            {"id": "https://abel/actor/alice", "preferredUsername": "alyssa"}
        )
        == "alyssa@abel"
    )



import pytest
from unittest.mock import AsyncMock

from . import Activity

remote_actor = "http://remote/actor"
local_actor = "http://local/actor"


@pytest.mark.parametrize(
    "activity",
    [
        {},
        {"type": "Undo"},
        {"type": "Undo", "object": "something"},
        {"type": "Undo", "object": {}},
        {"type": "Undo", "object": {"type": "Invite"}},
        {"type": "Undo", "object": {"type": "Follow", "id": "http://remote/follow"}},
        {
            "type": "Undo",
            "actor": local_actor,
            "object": {"type": "Follow", "actor": "http://local/actor/two"},
        },
    ],
)
async def test_undo_follow_bad_cases(activity):
    retrieve = AsyncMock()

    follow = await Activity(activity, domain="local").undo_of_follow(retrieve)

    assert follow is None


@pytest.mark.parametrize(
    "activity",
    [
        {
            "type": "Undo",
            "actor": local_actor,
            "object": {"type": "Follow", "object": remote_actor},
        },
        {
            "type": "Undo",
            "actor": {"id": local_actor},
            "object": {"type": "Follow", "object": remote_actor},
        },
        {
            "type": "Undo",
            "actor": local_actor,
            "object": {"type": "Follow", "object": {"id": remote_actor}},
        },
        {
            "type": "Undo",
            "actor": {"id": local_actor},
            "object": {"type": "Follow", "object": {"id": remote_actor}},
        },
    ],
)
async def test_accept_follow_good_case(activity):
    retrieve = AsyncMock()

    follow = await Activity(activity, domain="local").undo_of_follow(retrieve)

    assert isinstance(follow, Activity)

    retrieve.assert_not_awaited()


async def test_accept_follow_good_case_with_retrieve():
    retrieve = AsyncMock(
        return_value={
            "id": "http://local/follow",
            "type": "Follow",
            "object": "http://remote/actor",
        }
    )
    activity = {
        "type": "Undo",
        "actor": "http://local/actor",
        "object": "http://local/follow",
    }

    follow = await Activity(activity, domain="local").undo_of_follow(retrieve)

    assert isinstance(follow, Activity)

    retrieve.assert_awaited_once_with("http://local/follow")



from unittest.mock import AsyncMock

import aiohttp

from bovine.crypto.test import did_key, ed25519_key

from .moo_auth import MooAuthClient


async def test_moo_auth_client_get():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()

    client = MooAuthClient(session, did_key, ed25519_key)

    await client.get(url)

    session.get.assert_awaited_once()

    args = session.get.await_args

    assert args[0] == (url,)
    assert "headers" in args[1]
    headers = args[1]["headers"]

    assert headers.keys() == {
        "accept",
        "date",
        "host",
        "authorization",
        "x-moo-signature",
        "user-agent",
    }



import json
from unittest.mock import AsyncMock, MagicMock

import aiohttp

from bovine import BovineActor
from .clients.moo_auth import MooAuthClient
from bovine.activitystreams import OrderedCollection, OrderedCollectionPage
from bovine.crypto.test import private_key


async def test_activity_pub_client_get_collection_no_pages():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()

    actor = (
        BovineActor()
        .with_actor_id("actor_id")
        .with_http_signature(public_key_url, private_key, session=session)
    )

    text_mock = AsyncMock()
    session.get.return_value = MagicMock(aiohttp.ClientResponse)
    session.get.return_value.text = text_mock

    items = [{"id": j} for j in range(7)]

    builder = OrderedCollection(id="url", count=7, items=items)
    text_mock.return_value = json.dumps(builder.build())

    result = await actor.get_ordered_collection(url)
    session.get.assert_awaited_once()

    assert result["total_items"] == 7
    assert result["items"] == items


async def test_activity_pub_client_get_collection_pages():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()

    actor = (
        BovineActor()
        .with_actor_id("actor_id")
        .with_http_signature(public_key_url, private_key, session=session)
    )
    text_mock = AsyncMock()
    session.get.return_value = MagicMock(aiohttp.ClientResponse)
    session.get.return_value.text = text_mock
    session.get.return_value.status = 200

    items = [{"id": j} for j in range(23)]

    builder = OrderedCollection(id="url", count=23, first="first", last="last")
    page_1 = OrderedCollectionPage(
        id="page_1", items=items[:13], part_of="url_1", next="next_1"
    )
    page_2 = OrderedCollectionPage(
        id="page_2", items=items[13:20], part_of="url_1", next="next_2"
    )
    page_3 = OrderedCollectionPage(id="page_3", items=items[20:], part_of="url_1")

    text_mock.side_effect = [
        json.dumps(builder.build()),
        json.dumps(page_1.build()),
        json.dumps(page_2.build()),
        json.dumps(page_3.build()),
    ]

    result = await actor.get_ordered_collection(url)

    assert result["total_items"] == 23
    assert result["items"] == items

    text_mock.side_effect = [
        json.dumps(builder.build()),
        json.dumps(page_1.build()),
        json.dumps(page_2.build()),
        json.dumps(page_3.build()),
    ]

    result = await actor.get_ordered_collection(url, max_items=17)

    assert result["total_items"] == 23
    assert result["items"] == items[:20]


async def test_bovine_actor_get_tombstone():
    actor = BovineActor()
    actor.client = AsyncMock(MooAuthClient)
    result = AsyncMock()
    result.status = 410
    actor.client.get.return_value = result

    result = await actor.get("https://tombstone.example")

    assert isinstance(result, dict)
    assert result["type"] == "Tombstone"



import aiohttp

from bovine.utils.date import get_gmt_now

from .event_source import EventSource
from .utils import BOVINE_CLIENT_NAME


class BearerAuthClient:
    """Client for using Bearer authentication"""

    def __init__(self, session: aiohttp.ClientSession, bearer_key: str):
        self.session = session
        self.bearer_key = bearer_key

    async def get(self, url: str, headers: dict = {}):
        accept = "application/activity+json"
        date_header = get_gmt_now()

        headers["accept"] = accept
        headers["date"] = date_header
        headers["authorization"] = f"Bearer {self.bearer_key}"
        headers["user-agent"] = BOVINE_CLIENT_NAME

        return await self.session.get(url, headers=headers)

    async def post(
        self, url: str, body: str, headers: dict = {}, content_type: str | None = None
    ):
        accept = "application/activity+json"
        # LABEL: ap-s2s-content-type
        if content_type is None:
            content_type = "application/activity+json"
        date_header = get_gmt_now()

        headers["accept"] = accept
        headers["content-type"] = content_type
        headers["date"] = date_header
        headers["authorization"] = f"Bearer {self.bearer_key}"

        return await self.session.post(url, data=body, headers=headers)

    def event_source(self, url: str, headers: dict = {}) -> EventSource:
        """Returns an EventSource for the server sent events given by url"""

        date_header = get_gmt_now()
        accept = "text/event-stream"

        headers["accept"] = accept
        headers["date"] = date_header
        headers["authorization"] = f"Bearer {self.bearer_key}"
        headers["user-agent"] = BOVINE_CLIENT_NAME

        return EventSource(self.session, url, headers=headers)



import json
import logging
import re

import aiohttp

from .utils import BOVINE_CLIENT_NAME

logger = logging.getLogger(__name__)


async def lookup_with_webfinger(
    session: aiohttp.ClientSession, webfinger_url: str, params: dict
):
    async with session.get(
        webfinger_url, params=params, headers={"user-agent": BOVINE_CLIENT_NAME}
    ) as response:
        if response.status != 200:
            logger.warning(f"{params['resource']} not found using webfinger")
            return None
        text = await response.text()
        data = json.loads(text)

        if "links" not in data:
            return None

        links = data["links"]
        for entry in links:
            if "rel" in entry and entry["rel"] == "self":
                if re.match(r"application/.*json", entry["type"]):
                    return entry["href"]

    return None



from datetime import datetime, timedelta, timezone

from dateutil.parser import parse


def get_gmt_now() -> str:
    return datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")


def parse_gmt(date_string: str) -> datetime:
    return parse(date_string)


def check_max_offset_now(dt: datetime, minutes: int = 5) -> bool:
    now = datetime.now(tz=timezone.utc)

    if dt > now + timedelta(minutes=minutes):
        return False

    if dt < now - timedelta(minutes=minutes):
        return False

    return True



import json
import logging
import traceback

import aiohttp

from .utils import BOVINE_CLIENT_NAME

logger = logging.getLogger(__name__)


async def fetch_nodeinfo_document(session: aiohttp.ClientSession, domain: str) -> dict:
    wellknown_nodeinfo_url = f"https://{domain}/.well-known/nodeinfo"

    async with session.get(
        wellknown_nodeinfo_url,
        headers={"user-agent": BOVINE_CLIENT_NAME},
        timeout=60,
    ) as response:
        return await response.json()


async def fetch_nodeinfo20(session: aiohttp.ClientSession, url: str) -> dict | None:
    try:
        async with session.get(
            url, headers={"user-agent": BOVINE_CLIENT_NAME}
        ) as response:
            text = await response.text()
            return json.loads(text)

    except Exception as e:
        logger.error(str(e))
        for log_line in traceback.format_exc().splitlines():
            logger.error(log_line)
        return None



from unittest.mock import AsyncMock

import aiohttp

from bovine.crypto.test import private_key

from .signed_http import SignedHttpClient


async def test_activity_pub_client_get():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()

    client = SignedHttpClient(session, public_key_url, private_key)

    await client.get(url)

    session.get.assert_awaited_once()



from dataclasses import dataclass, field
from typing import Optional, Set

from .utils import id_for_object


@dataclass
class Activity:
    """A dataclass representing an `ActivityStreams Activity
    <https://www.w3.org/TR/activitystreams-vocabulary/#activity-types>`_"""

    type: str
    actor: Optional[str] = None
    followers: Optional[str] = None
    id: Optional[str] = None
    published: Optional[str] = None
    to: Set[str] = field(default_factory=set)
    cc: Set[str] = field(default_factory=set)

    name: Optional[str] = None
    summary: Optional[str] = None
    content: Optional[str] = None

    target: Optional[str] = None
    object: Optional[str] = None

    def as_public(self):
        """makes the activity public, i.e. public in to and followers in cc"""
        self.to.add("https://www.w3.org/ns/activitystreams#Public")
        if self.followers:
            self.cc.add(self.followers)
        return self

    def as_followers(self):
        """addresses the activity to followers, if they are set"""
        if self.followers:
            self.to.add(self.followers)
        return self

    def as_unlisted(self):
        """makes the activity unlisted, i.e. public in cc and followers in to"""
        if self.followers:
            self.to.add(self.followers)
        self.cc.add("https://www.w3.org/ns/activitystreams#Public")
        return self

    def build(self) -> dict:
        """converts the activity into a dict, that can be serialized to JSON"""
        result = {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": self.type,
            "actor": self.actor,
            "to": list(self.to),
            "cc": list(self.cc - self.to),
        }

        extra_fields = {
            "id": self.id,
            "published": self.published,
            "name": self.name,
            "summary": self.summary,
            "content": self.content,
            "target": self.target,
            "object": self.object,
        }

        if result["to"] is not None and len(result["to"]) == 0:
            del result["to"]
        if result["cc"] is not None and len(result["cc"]) == 0:
            del result["cc"]

        for key, value in extra_fields.items():
            if value:
                result[key] = value

        return result


class ActivityFactory:
    """Basic factory for Activity objects.

    Usally created by a BovineClient"""

    def __init__(self, actor_information):
        self.information = actor_information

    def create(self, obj, **kwargs):
        """Activity of type Create from Object"""
        return Activity(
            type="Create",
            actor=obj.get("attributedTo", self.information["id"]),
            cc=set(obj.get("cc", [])),
            to=set(obj.get("to", [])),
            object=obj,
            **kwargs,
        )

    def update(self, obj, **kwargs):
        """Activity of type Update from Object"""
        return Activity(
            type="Update",
            actor=obj.get("attributedTo", self.information["id"]),
            cc=set(obj.get("cc", [])),
            to=set(obj.get("to", [])),
            object=obj,
            **kwargs,
        )

    def like(self, target, **kwargs):
        """Like for target"""
        return Activity(
            type="Like", actor=self.information["id"], object=target, **kwargs
        )

    def delete(self, target, **kwargs):
        """Delete for target"""
        return Activity(
            type="Delete", actor=self.information["id"], object=target, **kwargs
        )

    def accept(self, obj, **kwargs):
        """Accept for object"""
        if isinstance(obj, str):
            return Activity(
                type="Accept", actor=self.information["id"], object=obj, **kwargs
            )
        return Activity(
            type="Accept",
            actor=self.information["id"],
            object=obj,
            to={id_for_object(obj.get("actor"))},
            **kwargs,
        )

    def reject(self, obj, **kwargs):
        """Reject for object"""
        return Activity(
            type="Reject",
            actor=self.information["id"],
            object=obj,
            to={id_for_object(obj.get("actor"))},
            **kwargs,
        )

    def announce(self, obj, **kwargs):
        """Announce for object"""
        return Activity(
            type="Announce",
            actor=self.information["id"],
            object=obj,
            followers=self.information.get("followers", []),
            **kwargs,
        )

    def follow(self, obj, **kwargs):
        """Follow for object"""

        return Activity(
            type="Follow",
            actor=self.information["id"],
            object=obj,
            to={id_for_object(obj)},
            **kwargs,
        )

    def undo(self, activity, **kwargs):
        """Undo for activity"""

        return Activity(
            type="Undo",
            actor=self.information["id"],
            object=activity,
            to={id_for_object(x) for x in activity.get("to")},
            **kwargs,
        )



from bovine.clients import lookup_uri_with_webfinger

from .bovine_object import BovineObject


class Actor(BovineObject):
    """Parses an actor object"""

    def __init__(self, data, domain: str = None, domain_may_differ: bool = False):
        super().__init__(data, domain=domain, domain_may_differ=domain_may_differ)

        if "outbox" not in self.data:
            raise ValueError("An actor must have an outbox")
        if not isinstance(self.data["outbox"], str):
            raise ValueError("The outbox must be a single string")
        if "inbox" not in self.data:
            raise ValueError("An actor must have an inbox")
        if not isinstance(self.data["inbox"], str):
            raise ValueError("The inbox must be a single string")
        self.webfinger_identifier = None

    @property
    def id(self):
        """id of the actor"""
        return self.data.get("id")

    @property
    def unvalidated_acct_uri(self) -> str | None:
        username = self.data.get("preferredUsername")
        if isinstance(username, dict):
            username = username.get("@value")
        if not isinstance(username, str):
            return None
        if len(username) == 0:
            return None
        return f"acct:{username}@{self.domain}"

    def publicKeyPem(self, key_id):
        key = self.data.get("publicKey", {})
        if key.get("owner") != self.id:
            raise Exception("Key has incorrect owner")
        return key.get("publicKeyPem")

    @property
    def identifiers(self):
        """Lists the identifiers of the actor.

        In order for the account uri to be listed as an identifier,
        the coroutine validate_acct_uri needs to be run"""
        result = set()
        if self.id:
            result.add(self.id)
        if self.webfinger_identifier:
            result.add(self.webfinger_identifier)
        return result

    async def validate_acct_uri(self, session=None) -> bool:
        """Checks if the acct uri defined by preferredUsername and
        the domain can be verified using a webfinger lookup.
        If yes, returns True and adds the acct uri to the
        identifiers of the actor."""
        lookup_id, _ = await lookup_uri_with_webfinger(
            session, self.unvalidated_acct_uri
        )

        if lookup_id == self.id:
            self.webfinger_identifier = self.unvalidated_acct_uri
            return True

        return False


class Activity(BovineObject):
    """Represents an activity"""

    @property
    def is_accept(self) -> bool:
        return self.data.get("type") == "Accept"

    @property
    def is_create(self) -> bool:
        return self.data.get("type") == "Create"

    @property
    def is_follow(self) -> bool:
        return self.data.get("type") == "Follow"

    @property
    def is_undo(self) -> bool:
        return self.data.get("type") == "Undo"

    async def object_for_create(self, retrieve) -> dict | None:
        """If activity is a create, returns the corresponding object"""
        if not self.is_create:
            return

        obj = self.data.get("object")
        if obj is None:
            return

        if isinstance(obj, str):
            obj = await retrieve(obj)
        if not isinstance(obj, dict):
            return
        obj = Object(obj, domain=self.domain)

        if (
            obj.attributed_to_id
            and self.actor_id
            and self.actor_id != obj.attributed_to_id
        ):
            raise ValueError("actor_id and attributed_to must match")

        return obj

    async def accept_for_follow(self, retrieve) -> dict | None:
        """If the activity is an Accept for a Follow request, returns said
        Follow request. Basic validation is run on the Accept and Follow
        request.

        In case of not an Accept for a Follow request or it being invalid
        None is returned.

        :param retrieve: a coroutine str -> dict that takes an object_id and
            resolves it to the corresponding object

        :param retrieve: a coroutine str -> dict that takes an object_id and
            resolves it to the corresponding object"""

        if not self.is_accept:
            return None
        follow = self.data.get("object")
        if not follow:
            return None
        if isinstance(follow, str):
            follow = await retrieve(follow)
        if not isinstance(follow, dict):
            return None

        follow = Activity(follow, domain=self.domain, domain_may_differ=True)

        if not follow.is_follow:
            return None

        if follow.object_id != self.actor_id:
            return None

        return follow

    async def undo_of_follow(self, retrieve) -> dict | None:
        """If the activity is an Undo of a Follow request, returns said
        Follow request. Basic validation is run on the Undo and Follow
        request.

        In case of not an Undo of a Follow request or it being invalid
        None is returned.

        :param retrieve: a coroutine str -> dict that takes an object_id and
            resolves it to the corresponding object"""

        if not self.is_undo:
            return None

        follow = self.data.get("object")
        if not follow:
            return None
        if isinstance(follow, str):
            follow = await retrieve(follow)
        if not isinstance(follow, dict):
            return None
        follow = Activity(follow, domain=self.domain, domain_may_differ=True)
        if not follow.is_follow:
            return None

        if follow.domain != self.domain:
            return None
        if follow.actor_id and self.actor_id and follow.actor_id != self.actor_id:
            return None

        return follow


class Object(BovineObject):
    """Basic representation of an object"""

    @property
    def attributed_to_id(self):
        attributed_to = self.data.get("attributedTo")
        if isinstance(attributed_to, dict):
            attributed_to = attributed_to.get("id")
        return attributed_to

    @property
    def in_reply_to(self):
        return self.data.get("inReplyTo")

    @property
    def identifier(self):
        return self.data.get("id")

    @property
    def mentions(self):
        return [x for x in self.tags if x.get("type") == "Mention"]

    @property
    def tags(self):
        tags = self.data.get("tag", [])
        if isinstance(tags, list):
            return tags
        return [tags]



from typing import Optional, Tuple

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, rsa
from multiformats import multibase, multicodec

import bovine

from .helper import content_digest_sha256, private_key_to_ed25519, public_key_to_did_key
from .http_signature import HttpSignature
from .signature_checker import SignatureChecker


def generate_ed25519_private_key() -> str:
    """Returns a multicodec/multibase encoded ed25519 private key"""
    private_key = ed25519.Ed25519PrivateKey.generate()

    private_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption(),
    )

    wrapped = multicodec.wrap("ed25519-priv", private_bytes)

    return multibase.encode(wrapped, "base58btc")


def private_key_to_did_key(private_key_str: str) -> str:
    """Computes public key in did key form of Ed25519 private key

    :param private_key_str: multibase/multicodec encoded Ed25519 private key

    :return: did:key"""

    private_key = private_key_to_ed25519(private_key_str)

    return public_key_to_did_key(private_key.public_key())


def generate_rsa_public_private_key() -> Tuple[str, str]:
    """Generates a new pair of RSA public and private keys.

    :returns: pem encoded public and private key
    """

    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    public_key = private_key.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    return public_key_pem.decode("utf-8"), private_key_pem.decode("utf-8")


def build_validate_http_signature(key_retriever):
    """Creates a validate_signature function. validate_signature takes the request
    as parameter and returns the owner if the http signature is valid.

    :param key_retriever:
        A coroutine that given a key id returns a tuple `public_key, owner`. Here
        `public_key` is assumed to be PEM encoded and owner is an URI. In the Fediverse
        use case, owner will be the actor id.

    Example

    .. code-block:: python

        async def retrieve(key_id):
            async with aiohttp.ClientSession() as session:
                response = await session.get(key_id)
                data = await response.json()
                key = data.get("publicKey", {})
                return key.get("publicKeyPem"), key.get("owner")

        validator = build_validate_http_signature(retrieve)

    `validator` then accepts as argument a `werzeug.wrappers.Request` object.
    """

    signature_checker = SignatureChecker(key_retriever)
    return signature_checker.validate_signature_request


def build_validate_http_signature_raw(key_retriever):
    """Creates a validate_signature function. validate_signature takes
    `(method, url, headers, body)` as parameters and returns
    the owner if the http signature is valid.
    The rest of behavior is as `build_validate_http_signature`.
    """

    signature_checker = SignatureChecker(key_retriever)
    return signature_checker.validate_signature


async def validate_moo_auth_signature(
    request, domain
) -> Tuple[Optional[str], Optional[str]]:
    """Validates the `Moo-Auth-1 <https://blog.mymath.rocks/2023-03-15/BIN1_Moo_Authentication_and_Authoriation>`_ signature of the request.
    Returns the did-key if the signature is valid.

    :param request:
        The request to validate the signature for.
    :param domain:
        The domain the request is made to.

    :returns:
        On success the did key and domain, on failure None, None
        When no domain is passed the did key and None is returned
    """  # noqa: E501
    didkey = request.headers["authorization"][11:]
    signature = request.headers["x-moo-signature"]

    dt = bovine.utils.date.parse_gmt(request.headers["date"])
    assert domain == request.headers["host"]
    assert bovine.utils.date.check_max_offset_now(dt)

    if request.method.lower() == "get":
        http_signature = (
            HttpSignature()
            .with_field("(request-target)", "get " + request.path)
            .with_field("host", request.headers["host"])
            .with_field("date", request.headers["date"])
        )
    else:
        raw_data = await request.get_data()
        digest = content_digest_sha256(raw_data)
        if digest != request.headers["digest"]:
            return None, None
        http_signature = (
            HttpSignature()
            .with_field("(request-target)", "post " + request.path)
            .with_field("host", request.headers["host"])
            .with_field("date", request.headers["date"])
            .with_field("digest", request.headers["digest"])
        )

    if http_signature.ed25519_verify(didkey, signature):
        return didkey, None
    return None, None



import aiohttp

from . import lookup_account_with_webfinger, lookup_did_with_webfinger, lookup_with_dns


async def test_lookup_account():
    async with aiohttp.ClientSession() as session:
        result = await lookup_account_with_webfinger(session, "helge@mymath.rocks")

    assert result.startswith("https://mymath.rocks/")


async def test_lookup_did_with_webfinger():
    did_helge = "did:key:z6MkujdZ216eYz55vz8X5HetqeJXj9ddn5ZHZUsBpRX4wfnL"
    async with aiohttp.ClientSession() as session:
        result = await lookup_did_with_webfinger(session, "mymath.rocks", did_helge)

    assert result.startswith("https://mymath.rocks/")


async def test_lookup_did_with_webfinger_with_protocol():
    did_helge = "did:key:z6MkujdZ216eYz55vz8X5HetqeJXj9ddn5ZHZUsBpRX4wfnL"
    async with aiohttp.ClientSession() as session:
        result = await lookup_did_with_webfinger(
            session, "https://mymath.rocks", did_helge
        )

    assert result.startswith("https://mymath.rocks/")


async def test_lookup_with_dns():
    domain = "mymath.rocks"

    async with aiohttp.ClientSession() as session:
        result = await lookup_with_dns(session, domain)

    assert (
        result
        == "https://mymath.rocks/endpoints/SYn3cl_N4HAPfPHgo2x37XunLEmhV9LnxCggcYwyec0"
    )


async def test_lookup_with_dns_no_result():
    domain = "duh.mymath.rocks"

    async with aiohttp.ClientSession() as session:
        result = await lookup_with_dns(session, domain)

    assert result is None



import pytest

from .signature import Signature


def test_signature_header_parsing():
    header_string = 'keyId="https://host.user#main-key",algorithm="rsa-sha256",headers="(request-target) host date digest content-type",signature="h...Kg=="'  # noqa E501

    result = Signature.from_signature_header(header_string)

    assert result.key_id == "https://host.user#main-key"
    assert result.algorithm == "rsa-sha256"
    assert result.headers == "(request-target) host date digest content-type"
    assert result.signature == "h...Kg=="
    assert result.fields == [
        "(request-target)",
        "host",
        "date",
        "digest",
        "content-type",
    ]


def test_signature_header_without_algorithm():
    header_string = 'keyId="https://host.user#main-key",headers="(request-target) host date",signature="stuff="'  # noqa E501

    result = Signature.from_signature_header(header_string)

    assert result.key_id == "https://host.user#main-key"
    assert result.algorithm == "rsa-sha256"
    assert result.headers == "(request-target) host date"
    assert result.signature == "stuff="
    assert result.fields == [
        "(request-target)",
        "host",
        "date",
    ]


def test_signature_header_unsupported_algorithm():
    header_string = 'keyId="https://host.user#main-key",algorithm="bad",headers="(request-target) host date",signature="stuff="'  # noqa E501

    with pytest.raises(ValueError):
        Signature.from_signature_header(header_string)



import logging
import json
import traceback
from urllib.parse import urlparse

import bovine.utils.date
from bovine.utils.date import parse_gmt

from .helper import content_digest_sha256
from .http_signature import HttpSignature
from .signature import parse_signature_header

logger = logging.getLogger(__name__)


class SignatureChecker:
    def __init__(self, key_retriever):
        self.key_retriever = key_retriever

    async def validate_signature(self, method, url, headers, body):
        if "signature" not in headers:
            logger.debug("Signature not present on request for %s", url)
            logger.debug(json.dumps(dict(headers)))
            return None

        if method.lower() == "post":
            digest = content_digest_sha256(await body())
        else:
            digest = None

        if digest is not None:
            request_digest = headers["digest"]
            request_digest = request_digest[:4].lower() + request_digest[4:]
            if request_digest != digest:
                logger.warning("Different digest")
                return None

        try:
            http_signature = HttpSignature()
            parsed_signature = parse_signature_header(headers["signature"])
            signature_fields = parsed_signature.fields

            if (
                "(request-target)" not in signature_fields
                or "date" not in signature_fields
            ):
                logger.warning("Required field not present in signature")
                return None

            if digest is not None and "digest" not in signature_fields:
                logger.warning("Digest not present, but computable")
                return None

            http_date = parse_gmt(headers["date"])
            if not bovine.utils.date.check_max_offset_now(http_date):
                logger.warning(f"Encountered invalid http date {headers['date']}")
                return None

            for field in signature_fields:
                if field == "(request-target)":
                    method = method.lower()
                    parsed_url = urlparse(url)
                    path = parsed_url.path
                    http_signature.with_field(field, f"{method} {path}")
                else:
                    http_signature.with_field(field, headers[field])

            public_key, owner = await self.key_retriever(parsed_signature.key_id)

            if public_key is None:
                logger.debug(f"Could not retrieve key from {parsed_signature.key_id}")
                return None

            if http_signature.verify(public_key, parsed_signature.signature):
                return owner

        except Exception as e:
            logger.error(str(e))
            logger.error(headers)
            for log_line in traceback.format_exc().splitlines():
                logger.error(log_line)
            return None

    async def validate_signature_request(self, request):
        return await self.validate_signature(
            request.method, request.url, request.headers, request.get_data
        )



from unittest.mock import AsyncMock, MagicMock

import aiohttp

from bovine.crypto.signature_checker import SignatureChecker
from bovine.crypto.test import private_key, public_key

from .signed_http_methods import signed_get


async def test_signed_get():
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()
    session.get.return_value = "value"

    key_retriever = AsyncMock()
    key_retriever.return_value = public_key, "owner"
    signature_checker = SignatureChecker(key_retriever)

    response = await signed_get(session, public_key_url, private_key, url)

    session.get.assert_awaited_once()

    assert response == "value"

    args = session.get.await_args

    assert args[0] == (url,)
    assert "headers" in args[1]
    headers = args[1]["headers"]

    assert headers.keys() == {"accept", "date", "host", "signature", "user-agent"}

    assert headers["accept"] == "application/activity+json"
    assert headers["host"] == "test_domain"

    request = MagicMock()
    request.headers = headers
    request.method = "get"
    request.url = url

    assert (await signature_checker.validate_signature_request(request)) == "owner"

    key_retriever.assert_awaited_once()
    assert key_retriever.await_args[0] == (public_key_url,)



from cryptography.hazmat.primitives.asymmetric import ed25519

from .helper import (
    content_digest_sha256,
    did_key_to_public_key,
    sign_message,
    verify_signature,
)
from .test import private_key, public_key


def test_content_digest_sha256():
    digest = content_digest_sha256("content")

    assert digest == "sha-256=7XACtDnprIRfIjV9giusFERzD722AW0+yUMil7nsn3M="


def test_did_to_public_key():
    did_example = "did:key:z6MkiTBz1ymuepAQ4HEHYSF1H8quG5GLVVQR3djdX3mDooWp"

    public_key = did_key_to_public_key(did_example)

    assert isinstance(public_key, ed25519.Ed25519PublicKey)


def test_crypto_sign_verify():
    message = "secret"

    signature = sign_message(private_key, message)

    assert verify_signature(public_key, message, signature)


def test_crypto_sign_verify_failure():
    message = "secret"

    assert not verify_signature(public_key, message, "")



import pytest

from .validation import validate_to, validate_tos


def test_validate_tos_none() -> None:
    assert validate_tos(None)


def test_validate_tos() -> None:
    assert validate_tos(["alysssa@ben.de"])


@pytest.mark.parametrize(
    "recipient", ["https://chatty.example.com/alyssa", "alyssa@chatty.com"]
)
def test_validate_to_true(recipient: str) -> None:
    assert validate_to(recipient)


@pytest.mark.parametrize(
    "recipient", ["http://chatty.example.com/alyssa", "alyssachatty.com"]
)
def test_validate_to_false(recipient: str) -> None:
    assert not validate_to(recipient)



import aiohttp
import json

from urllib.parse import urlparse
from urllib.robotparser import RobotFileParser
from dataclasses import dataclass
from bs4 import BeautifulSoup
from functools import cached_property
from bovine.activitystreams.object_factory import Object
from bovine.clients.utils import BOVINE_CLIENT_NAME
from bovine.jsonld import with_activitystreams_context


def safe_json_loads(text: str) -> dict | None:
    try:
        return json.loads(text)
    except Exception:
        return None


class RobotFileDeniesAccess(Exception):
    """Used to indicate that robots.txt does not allow the user agent
    to access the url being queried"""

    pass


@dataclass
class WebPage:
    """Class to capture loading webpages and transforming their
    content in objects more usable in the Fediverse."""

    url: str
    text: str | None = None

    async def fetch(self, session: aiohttp.ClientSession | None = None):
        """Fetches the webpage and transform its content using
        BeautifulSoup"""
        if session is None:
            async with aiohttp.ClientSession() as session:
                await self._fetch_with_session(session)
        else:
            await self._fetch_with_session(session)

    async def _fetch_with_session(self, session):
        async with session.get(
            self.robots_url, headers={"user-agent": BOVINE_CLIENT_NAME}
        ) as response:
            robots = RobotFileParser()
            robots.parse((await response.text()).split("\n"))

            if not robots.can_fetch(BOVINE_CLIENT_NAME, self.url):
                raise RobotFileDeniesAccess()
        async with session.get(
            self.url, headers={"accept": "text/html", "user-agent": BOVINE_CLIENT_NAME}
        ) as response:
            self.text = await response.text()

    @cached_property
    def soup(self):
        return BeautifulSoup(self.text, features="lxml")

    @cached_property
    def jsonld(self) -> dict | list:
        """Usage

        .. code-block:: python

            from bovine.clients.web import WebPage

            page = WebPage(
                "https://www.allrecipes.com/recipe/263822/pasta-alla-norma-eggplant-pasta/"
            )

            print(page.jsonld[0][0])
        """
        raw = self.soup.find_all("script", attrs={"type": "application/ld+json"})

        raw = [safe_json_loads(tag.text) for tag in raw]

        return [x for x in raw if x]

    def meta_content_for_property(self, value: str) -> str | None:
        tag = self.soup.find("meta", attrs={"property": value})
        if tag:
            return tag.get("content")
        return None

    def meta_content_for_property_int(self, value: str) -> int | None:
        tag = self.soup.find("meta", attrs={"property": value})
        if tag:
            return int(tag.get("content"))
        return None

    @cached_property
    def open_graph_page(self) -> dict:
        """Creates an ActivityPub Page object from the Open Graph data"""
        image = Object(
            type="Image",
            url=self.meta_content_for_property("og:image"),
            name=self.meta_content_for_property("og:image:alt"),
            height=self.meta_content_for_property_int("og:image:height"),
            width=self.meta_content_for_property_int("og:image:width"),
            media_type=self.meta_content_for_property("og:image:type"),
        )
        page = Object(
            type="Page",
            name=self.meta_content_for_property("og:title"),
            url=self.meta_content_for_property("og:url"),
            summary=self.meta_content_for_property("og:description"),
            source={"url": self.url, "mediaType": "text/html"},
        )

        page.icon = image.build()

        return with_activitystreams_context(page.build())

    @cached_property
    def robots_url(self):
        parsed = urlparse(self.url)
        return f"{parsed.scheme}://{parsed.netloc}/robots.txt"



import asyncio
import json
from argparse import ArgumentParser

from asyncstdlib.itertools import islice
from ptpython.repl import embed

from bovine import BovineClient
from bovine.activitystreams.utils.print import print_activity


def build_parser():
    parser = ArgumentParser("Opens a REPL with preloaded BovineClient client")
    parser.add_argument("--domain", help="Domain the actor can be found on")
    parser.add_argument("--secret", help="Secret associated with the account")
    parser.add_argument("--config_file", help="Toml fail containing domain and secret")

    parser.add_argument("--inbox", action="store_true", help="Display the inbox")
    parser.add_argument("--outbox", action="store_true", help="Display the outbox")
    parser.add_argument(
        "--summary",
        action="store_true",
        help="Display a summary of the activity instead of the full json",
    )

    parser.add_argument(
        "--max_number",
        type=int,
        help="Number of elements to display in inbox or outbox",
        default=20,
    )

    return parser


def config(repl):
    repl.use_code_colorscheme("dracula")
    repl.enable_output_formatting = True


async def repl(client):
    async with client:
        activity_factory, object_factory = client.factories
        print("The variable client contains your BovineClient")
        print("The variables activity_factory and object_factory")
        print("contain the corresponding objects")
        print("With await client.inbox() and await client.outbox()")
        print("one can interface with these two")
        print()
        await embed(
            globals=globals(),
            locals=locals(),
            return_asyncio_coroutine=True,
            patch_stdout=True,
            configure=config,
        )


async def show_inbox(client: BovineClient, max_number: int = 10, summary: bool = False):
    async with client:
        await display_box(await client.inbox(), max_number, summary)


async def show_outbox(
    client: BovineClient, max_number: int = 10, summary: bool = False
):
    async with client:
        await display_box(await client.outbox(), max_number, summary)


async def display_box(box, max_number: int, summary: bool):
    async for item in islice(box, max_number):
        if summary:
            print_activity(item)
        else:
            print(json.dumps(item, indent=2))
            print()


def main():
    args = build_parser().parse_args()

    if args.config_file:
        client = BovineClient.from_file(args.config_file)
    elif args.domain and args.secret:
        client = BovineClient(host=args.domain, secret=args.secret)
    else:
        default_config_file = "bovine_user.toml"
        client = BovineClient.from_file(default_config_file)

        print(f"Config file not specified using fallback value '{default_config_file}'")

    if args.inbox:
        asyncio.run(
            show_inbox(client, max_number=args.max_number, summary=args.summary)
        )
    elif args.outbox:
        asyncio.run(
            show_outbox(client, max_number=args.max_number, summary=args.summart)
        )
    else:
        loop = asyncio.new_event_loop()
        loop.run_until_complete(repl(client))


if __name__ == "__main__":
    main()



import argparse
import asyncio

from .utils.msg import process, validate


def build_parser():
    parser = argparse.ArgumentParser("Send a message through your ActivityPub Actor")
    parser.add_argument(
        "--secret",
        help="Secret corresponding to a did-key deposited with your actor",
    )
    parser.add_argument("--host", help="Hostname your actor is on")
    parser.add_argument(
        "--to",
        nargs="+",
        help="Recipients either as FediVerse handle or actor id",
        default=[],
    )
    parser.add_argument("--public", action="store_true", default=False)
    parser.add_argument("message", help="You message text")

    return parser


def main():
    args = (build_parser()).parse_args()

    validate(args)

    asyncio.run(process(args))


if __name__ == "__main__":
    main()



from bovine.crypto.helper import content_digest_sha256
from bovine.crypto.http_signature import build_signature
from bovine.utils.date import get_gmt_now

from .event_source import EventSource
from .utils import BOVINE_CLIENT_NAME, host_target_from_url


class MooAuthClient:
    """Client for using Moo-Auth-1 authentication"""

    def __init__(self, session, did_key, private_key):
        self.session = session
        self.did_key = did_key
        self.private_key = private_key

    async def get(self, url, headers={}):
        host, target = host_target_from_url(url)

        accept = "application/activity+json"
        date_header = get_gmt_now()

        signature_helper = build_signature(host, "get", target).with_field(
            "date", date_header
        )
        signature_header = signature_helper.ed25519_sign(self.private_key)

        headers = {
            "accept": accept,
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            **signature_helper.headers,
            "authorization": f"Moo-Auth-1 {self.did_key}",
            "x-moo-signature": signature_header,
        }

        return await self.session.get(url, headers=headers)

    async def post(self, url, body, headers={}, content_type=None):
        host, target = host_target_from_url(url)
        accept = "application/activity+json"
        # LABEL: ap-s2s-content-type
        if content_type is None:
            content_type = "application/activity+json"
        date_header = get_gmt_now()

        digest = content_digest_sha256(body)

        signature_helper = (
            build_signature(host, "post", target)
            .with_field("date", date_header)
            .with_field("digest", digest)
        )
        signature_header = signature_helper.ed25519_sign(self.private_key)

        headers = {
            "accept": accept,
            "content-type": content_type,
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            **signature_helper.headers,
            "authorization": f"Moo-Auth-1 {self.did_key}",
            "x-moo-signature": signature_header,
        }

        return await self.session.post(url, data=body, headers=headers)

    def event_source(self, url, headers={}):
        host, target = host_target_from_url(url)
        date_header = get_gmt_now()
        accept = "text/event-stream"
        signature_helper = build_signature(host, "get", target).with_field(
            "date", date_header
        )
        signature_header = signature_helper.ed25519_sign(self.private_key)

        headers = {
            "accept": accept,
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            **signature_helper.headers,
            "authorization": f"Moo-Auth-1 {self.did_key}",
            "x-moo-signature": signature_header,
        }

        return EventSource(self.session, url, headers=headers)



def validate_tos(tos: list | None):
    if tos is None:
        return True
    return all(validate_to(to) for to in tos)


def validate_to(to: str):
    if to.startswith("https://"):
        return True

    if "@" in to:
        return True

    return False



import json
import logging
from typing import Optional
from urllib.parse import urlencode, urlparse

import tomllib

from .activitystreams import Collection
from .activitystreams.activity_factory import ActivityFactory
from .activitystreams.collection_helper import CollectionHelper
from .activitystreams.object_factory import ObjectFactory, Object
from .clients.authorization_wrapper import AuthorizationWrapper
from .utils import activity_pub_object_id_from_html_body

logger = logging.getLogger(__name__)


class BovineClient(AuthorizationWrapper):
    """BovineClient is meant to serve as the basis of building ActivityPub Clients.
    It defines methods for interacting with the endpoints defined by the corresponding
    ActivityPub Actor: inbox, outbox, and proxyUrl.

    Usage is either:

    .. code-block:: python

        async with BovineClient(**config) as actor:
            await do_something(actor)

    or

    .. code-block:: python

        actor = await BovineClient(**config)
        await actor.init()
        await do_something(actor)

    I still call the variable actor as it represents the ActivityPub Actor through
    a client.

    :param domain: Specifies the domain the actor is on, used for Moo-Auth-1
    :param host: Alias for domain
    :param actor_id: URI of the actor, used for Bearer Auth and HTTP Signatures
    :param account_url: Alias for actor_id
    :param secret: The private key material for Moo-Auth-1 and HTTP Signatures
    :param access_token: The access token for Bearer auth
    """

    def __init__(self, **kwargs):
        super().__init__(
            domain=kwargs.get("domain", kwargs.get("host")),
            secret=kwargs.get("secret"),
            actor_id=kwargs.get("actor_id", kwargs.get("account_url")),
            public_key_url=kwargs.get("public_key_url"),
            access_token=kwargs.get("access_token"),
        )

        self.information: Optional[dict] = None
        self._activity_factory = None
        self._object_factory = None

    async def get(self, target):
        assert self.client

        response = await self.client.get(target)
        response.raise_for_status()
        return json.loads(await response.text())

    async def init(self, session=None):
        """Manually initializes the BovineClient for cases when
        not used within async with. Also loads the actor information.

        :param session:
            can be used to specify an existing aiohttp.ClientSession. Otherwise a new
            one is created.
        """

        await super().init(session=session)

        if self.client is None:
            raise Exception("Client not set in BovineClient")
        self.information = await self.get(self.actor_id)

        logger.debug("Retrieved information %s", self.information)

        if any(required not in self.information for required in ["inbox", "outbox"]):
            raise Exception("Retrieved incomplete actor data")

    async def send_to_outbox(self, data: dict):
        """sends data to outbox of actor

        :param data: The data to send as python dict

        :return:
            The aiohttp.ClientResponse object. This means
            return_value.headers["location"] will contain the id of the
            posted activity.
        """
        if self.information is None:
            await self.init()

        assert self.client

        return await self.client.post(self.information["outbox"], json.dumps(data))

    async def proxy(self, target: str):
        """Retrieve's an element through the actors' proxyUrl endpoint
        as specified in ActivityPub.

            :param target: The URL of the object to retrieve


        FIXME: Support for non-json stuff"""
        response = await self.client.post(
            self.information["endpoints"]["proxyUrl"],
            urlencode({"id": target}),
            content_type="application/x-www-form-urlencoded",
        )
        response.raise_for_status()
        return await response.json()

    async def event_source(self):
        """Returns an EventSource corresponding to the actor's

        The syntax for this will probably change"""
        if self.information is None:
            await self.load()

        event_source_url = self.information["endpoints"]["eventSource"]
        return self.client.event_source(event_source_url)

    async def simplify_collection(self, collection):
        """Returns a Collection containing all items from the passed collection
        or collection id"""
        # items = await all_collection_elements(self, collection)

        items = []

        if isinstance(collection, str):
            collection_id = collection
        else:
            collection_id = collection.get("id")
        return Collection(id=collection_id, items=items).build()

    @property
    def activity_factory(self):
        """Returns an ActivityFactory for objects corresponding to the client's actor"""
        if self._activity_factory is None:
            self._activity_factory = ActivityFactory(self.information)
        return self._activity_factory

    @property
    def object_factory(self):
        """Returns an ObjectFactory for objects corresponding to the client's actor"""
        if self._object_factory is None:
            self._object_factory = ObjectFactory(client=self)
        return self._object_factory

    @property
    def factories(self):
        return self.activity_factory, self.object_factory

    @property
    def host(self):
        """The host the actor is on"""
        return urlparse(self.actor_id).netloc

    @property
    def followers(self) -> str:
        """The id of the follows collection"""
        return self.information["followers"]

    def inbox(self, resolve=True):
        """Provides a CollectionHelper for the Actors inbox"""
        inbox_collection = CollectionHelper(
            self.information["inbox"], self, resolve=resolve
        )
        return inbox_collection

    def outbox(self, resolve=True):
        """Provides a CollectionHelper for the Actors outbox"""
        outbox_collection = CollectionHelper(
            self.information["outbox"], self, resolve=resolve
        )
        return outbox_collection

    def collection_helper(self, collection, resolve=False):
        """Returns a CollectionHelper for the collection provided. Usage:

        .. code-block:: python

            async for x in client.collection_helper(uri_of_collection):
                await do_something(x)

        :param collection: Uri of the collection to irater over
        :param resolve: If true objects are automatically fetched"""

        return CollectionHelper(collection, self, resolve=resolve)

    @staticmethod
    def from_file(config_file: str):
        """Initializes the BovineClient from a toml config file"""
        with open(config_file, "rb") as fp:
            config = tomllib.load(fp)

        return BovineClient(**config)


class BovineActor(AuthorizationWrapper):
    r"""Defines the Bovine version of an ActivityPub Actor. This class is meant
    to be used when implementing an ActivityPub Server in order to handle the
    HTTP requests to another server.

    Currently most of these interactions use HTTP Signatures.

    Usage is either:

    .. code-block:: python

        async with BovineActor(**config) as actor:
            await do_something(actor)

    or

    .. code-block:: python

        actor = await BovineActor(**config)
        await actor.init()
        await do_something(actor)

    :param domain: Specifies the domain the actor is on, used for Moo-Auth-1
    :param host: Alias for domain
    :param actor_id: URI of the actor, used for Bearer Auth and HTTP Signatures
    :param account_url: Alias for actor_id
    :param secret: The private key material for Moo-Auth-1 and HTTP Signatures
    :param access_token: The access token for Bearer auth
    """

    def __init__(self, **config):
        super().__init__(
            domain=config.get("domain", config.get("host")),
            secret=config.get("secret"),
            actor_id=config.get("actor_id", config.get("acccount_url")),
            public_key_url=config.get("public_key_url"),
            access_token=config.get("access_token"),
        )

    async def init(self, session=None):
        """Manually initializes the BovineActor for cases when not used
        within async with

        :param session:
            can be used to specify an existing aiohttp.ClientSession. Otherwise a new
            one is created.
        """

        await super().init(session=session)

    async def post(self, target: str, data: dict):
        """Send a signed post with data to target"""
        response = await self.client.post(target, json.dumps(data))
        response.raise_for_status()

        return response

    async def get(self, target: str, fail_silently: bool = False):
        """Retrieve target with a get. An exception is raised if the request fails

        :param target: The URL of the object to retrieve
        :param fail_silently: do not raise an exception if the request fails
        """
        response = await self.client.get(target)

        if response.status >= 300 and response.status < 400:
            if "Location" in response.headers:
                return await self.get(response.headers["Location"])
            return None

        if response.status in [404, 410]:
            try:
                body = await response.text()
                result = json.loads(body)
                if result["type"] == "Tombstone":
                    return result
            except Exception:
                ...

            return Object(type="Tombstone", id=target).build()
        if fail_silently and response.status >= 400:
            return None

        response.raise_for_status()

        body = await response.text()

        try:
            return json.loads(body)
        except Exception:
            # if "json" not in response.headers.get("Content-Type"):
            if "alternate" in response.links:
                link_header = response.links["alternate"]
                if "json" in link_header.get("type", ""):
                    return await self.get(
                        str(link_header["url"]), fail_silently=fail_silently
                    )

            object_id = activity_pub_object_id_from_html_body(body)

            if object_id:
                return await self.get(object_id, fail_silently=fail_silently)
            else:
                return None

    async def get_ordered_collection(self, url: str, max_items: Optional[int] = None):
        """Retrieve target ordered collection

        :param url: url of the ordered collection
        :param max_items: maximal number of items to retrieve, use None for all
        """
        result = await self.client.get(url)
        result.raise_for_status()

        data = json.loads(await result.text())

        total_number_of_items = data["totalItems"]
        items = []

        if "orderedItems" in data:
            items = data["orderedItems"]

        if len(items) == total_number_of_items:
            return {"total_items": total_number_of_items, "items": items}

        if "first" in data:
            page_data = await self.get(data["first"])

            items = page_data["orderedItems"]

            while "next" in page_data and len(page_data["orderedItems"]) > 0:
                if max_items and len(items) > max_items:
                    return {"total_items": total_number_of_items, "items": items}

                page_data = await self.get(page_data["next"])

                items += page_data["orderedItems"]

        return {"total_items": total_number_of_items, "items": items}

    @staticmethod
    def from_file(config_file: str):
        with open(config_file, "rb") as fp:
            config = tomllib.load(fp)

        return BovineActor(**config)



import re

from . import now_isoformat, webfinger_response_json


def test_webfinger():
    response = webfinger_response_json("acct:name@domain", "url")

    assert "subject" in response
    assert "links" in response


def test_now_isoformat():
    result = now_isoformat()
    assert re.match(
        r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$",
        result,
    )



from . import WebPage


def test_determine_robots_url():
    page = WebPage(
        "https://www.allrecipes.com/recipe/263822/pasta-alla-norma-eggplant-pasta/"
    )

    assert page.robots_url == "https://www.allrecipes.com/robots.txt"


async def test_web_page():
    page = WebPage(
        "https://www.allrecipes.com/recipe/263822/pasta-alla-norma-eggplant-pasta/"
    )

    page.text = """
<html>
<head>
<script type="application/ld+json">{"@id": "https://jsonld.example"}</script>
</head>
</html>
"""

    jsonld = page.jsonld

    assert jsonld == [{"@id": "https://jsonld.example"}]


async def test_open_graph_page():
    page = WebPage("https://www.imdb.com/title/tt0117500/")
    page.text = """
<html prefix="og: https://ogp.me/ns#">
<head>
<title>The Rock (1996)</title>
<meta property="og:title" content="The Rock" />
<meta property="og:type" content="video.movie" />
<meta property="og:url" content="https://www.imdb.com/title/tt0117500/" />
<meta property="og:image" content="https://ia.media-imdb.com/images/rock.jpg" />
<meta property="og:description" 
  content="Sean Connery found fame and fortune as the suave, sophisticated British agent, James Bond." />
<meta property="og:image:type" content="image/jpeg" />
<meta property="og:image:width" content="400" />
<meta property="og:image:height" content="300" />
<meta property="og:image:alt" content="A shiny red apple with a bite taken out" />
</head>
</html>
"""

    result = page.open_graph_page

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "type": "Page",
        "name": "The Rock",
        "summary": "Sean Connery found fame and fortune as the suave, sophisticated British agent, James Bond.",
        "url": "https://www.imdb.com/title/tt0117500/",
        "icon": {
            "type": "Image",
            "mediaType": "image/jpeg",
            "url": "https://ia.media-imdb.com/images/rock.jpg",
            "name": "A shiny red apple with a bite taken out",
            "height": 300,
            "width": 400,
        },
        "source": {
            "url": "https://www.imdb.com/title/tt0117500/",
            "mediaType": "text/html",
        },
    }



import pytest
from unittest.mock import AsyncMock

from . import Activity, Object

remote_actor = "http://remote/actor"


@pytest.mark.parametrize(
    "activity",
    [
        {},
        {"type": "Create"},
        {"type": "Create", "object": "something"},
    ],
)
async def test_create_bad_cases(activity):
    retrieve = AsyncMock()

    obj = await Activity(activity, domain="remote").object_for_create(retrieve)

    assert obj is None


@pytest.mark.parametrize(
    "activity",
    [
        {
            "type": "Create",
            "actor": remote_actor,
            "object": {"id": "http://local/object"},
        },
        {
            "type": "Create",
            "actor": remote_actor,
            "object": {
                "id": "http://remote/object",
                "attributedTo": "http://remote/other",
            },
        },
    ],
)
async def test_create_bad_cases_with_exception(activity):
    retrieve = AsyncMock()

    with pytest.raises(ValueError):
        await Activity(activity, domain="remote").object_for_create(retrieve)


@pytest.mark.parametrize(
    "activity",
    [
        {
            "type": "Create",
            "object": {},
        },
        {
            "type": "Create",
            "object": {"type": "Note", "attributedTo": remote_actor},
        },
        {
            "type": "Create",
            "actor": remote_actor,
            "object": {"type": "Note"},
        },
        {
            "type": "Create",
            "actor": remote_actor,
            "object": {"type": "Note", "attributedTo": remote_actor},
        },
        {
            "type": "Create",
            "actor": {"id": remote_actor},
            "object": {"type": "Note", "attributedTo": remote_actor},
        },
        {
            "type": "Create",
            "actor": remote_actor,
            "object": {"type": "Note", "attributedTo": {"id": remote_actor}},
        },
        {
            "type": "Create",
            "actor": {"id": remote_actor},
            "object": {"type": "Note", "attributedTo": {"id": remote_actor}},
        },
    ],
)
async def test_create_good_cases(activity):
    retrieve = AsyncMock()

    obj = await Activity(activity, domain="remote").object_for_create(retrieve)

    assert isinstance(obj, Object)

    retrieve.assert_not_awaited()


async def test_create_with_retrieving_remote_object():
    obj_id = remote_actor + "/object"
    retrieve = AsyncMock(
        return_value={
            "id": obj_id,
            "type": "Note",
            "attributedTo": remote_actor,
            "content": "note",
        }
    )
    activity = {"type": "Create", "actor": remote_actor, "object": obj_id}

    obj = await Activity(activity, domain="remote").object_for_create(retrieve)

    assert isinstance(obj, Object)

    retrieve.assert_awaited_once_with(obj_id)



from . import JrdData, JrdLink


def test_jrd_link():
    assert JrdLink().build() == {}
    assert JrdLink(rel="rel", href="href", type="type").build() == {
        x: x for x in ["rel", "href", "type"]
    }


def test_jrd_asdict():
    assert JrdData().build() == {}
    assert JrdData(subject="acct:test").build() == {"subject": "acct:test"}

    link = JrdLink(href="https://some.example/")

    assert JrdData(links=[link]).build() == {
        "links": [{"href": "https://some.example/"}]
    }



import aiohttp
import bovine.clients.signed_http_methods


class SignedHttpClient:
    """Client for using HTTP Signatures"""

    def __init__(self, session, public_key_url, private_key, account_url=None):
        self.session = session
        self.public_key_url = public_key_url
        self.private_key = private_key
        self.account_url = account_url

    def set_session(self, session):
        self.session = session

        return self

    async def get(self, url, headers={}) -> aiohttp.ClientResponse:
        """Retrieves url using a signed get request"""
        return await bovine.clients.signed_http_methods.signed_get(
            self.session, self.public_key_url, self.private_key, url, headers
        )

    async def post(self, url, body, headers={}, content_type=None):
        """Posts to url using a signed post request"""
        return await bovine.clients.signed_http_methods.signed_post(
            self.session,
            self.public_key_url,
            self.private_key,
            url,
            body,
            headers,
            content_type=content_type,
        )

    def event_source(self, url):
        return bovine.clients.signed_http_methods.signed_event_source(
            self.session, self.public_key_url, self.private_key, url
        )



from importlib.metadata import version
from urllib.parse import urlparse

BOVINE_CLIENT_NAME = "bovine/" + version("bovine")


def host_target_from_url(url):
    parsed_url = urlparse(url)

    path = parsed_url.path
    if parsed_url.query:
        path = f"{path}?{parsed_url.query}"
    return parsed_url.netloc, path



from dataclasses import dataclass, asdict, field
from datetime import datetime, timezone
from typing import Optional, Tuple

from bs4 import BeautifulSoup


@dataclass
class JrdLink:
    """Dataclass representing a JRD link

    See `RFC 7033 <https://www.rfc-editor.org/rfc/rfc7033.html>`_
    """

    rel: str | None = None
    href: str | None = None
    type: str | None = None

    def build(self):
        """Transforms the JrdLink into a dict"""
        return {key: value for key, value in asdict(self).items() if value}


@dataclass
class JrdData:
    """Dataclass representing a JRD Object

    See `RFC 7033 <https://www.rfc-editor.org/rfc/rfc7033.html>`_
    """

    subject: str | None = None
    links: list[JrdLink] = field(default_factory=list)

    def build(self):
        """Transforms JrdData to dict"""
        result = {}
        if self.subject:
            result["subject"] = self.subject

        if len(self.links) > 0:
            result["links"] = [x.build() for x in self.links]

        return result


def webfinger_response_json(account: str, url: str) -> dict:
    """helper to generate a webfinger response"""
    return JrdData(
        subject=account,
        links=[JrdLink(href=url, rel="self", type="application/activity+json")],
    ).build()


def parse_fediverse_handle(account: str) -> Tuple[str, Optional[str]]:
    """Splits fediverse handle in name and domain"""
    if account[0] == "@":
        account = account[1:]

    if "@" in account:
        user, domain = account.split("@", 1)
        return user, domain
    return account, None


def now_isoformat() -> str:
    """Returns now in Isoformat, e.g. "2023-05-31T18:11:35Z", to be used as the value
    of published"""
    return (
        datetime.now(tz=timezone.utc).replace(microsecond=0, tzinfo=None).isoformat()
        + "Z"
    )


def activity_pub_object_id_from_html_body(body: str) -> str | None:
    """Determines the object identifier from the html body
    by parsing it and looking for link tags with rel="alternate"
    and type application/activity+json"""

    soup = BeautifulSoup(body, features="lxml")
    element = soup.find(
        "link", attrs={"rel": "alternate", "type": "application/activity+json"}
    )
    if not element:
        return None

    return element.attrs.get("href")
